@@ -8,29 +8,52 @@
 #include "gmem.h"
 #include "cover.h"
 
-#define CHAR_LINE 4
+#define CHAR_LINE 2
 
 /* How big will the initial bit set allocation be. */
-#define COVER_INITIAL_SIZE 16   /* 8 * CHAR_BIT = 64 bits (lines) */
+#define COVER_INITIAL_SIZE 16   /* 16 * CHAR_BIT = 128 bits (lines) */
 
 #define COVER_LIST_INITIAL_SIZE 8   /* 8 files in the hash */
 
-/* Handle an array of unsigned char as a bit set. */
-#define LINE_SET_COVERED(data, line)   data[line/CHAR_LINE] |=  (1 << (line%CHAR_LINE))
-#define LINE_SET_PRESENT(data, line)   data[line/CHAR_LINE] |=  (16 << (line%CHAR_LINE))
-#define LINE_IS_COVERED(data, line)    (data[line/CHAR_LINE] &   (1 << (line%CHAR_LINE)))
-#define LINE_IS_PRESENT_OR_COVERED(data, line)    (data[line/CHAR_LINE] &   (17 << (line%CHAR_LINE)))
+/* Handle an array of unsigned char as an array of 4 bit values per line,
+ * where bits 0-2 encode coverage and compiler phase (PL_phase has numeric
+ * values from 0 to 6), and bit 3 is presence flag.
+ */
+#define LINE_SHIFT(line)                      ((line%CHAR_LINE)*4)
+#define LINE_SET_COVERED(data, line, phase)     data[line/CHAR_LINE] |= ((phase+1) << LINE_SHIFT(line))
+#define LINE_SET_PRESENT(data, line)            data[line/CHAR_LINE] |= (8 << LINE_SHIFT(line))
+#define LINE_IS_COVERED(data, line)           ((data[line/CHAR_LINE] & ( 7 << LINE_SHIFT(line))) != 0)
+#define LINE_IS_PRESENT_OR_COVERED(data, line) (data[line/CHAR_LINE] & (15 << LINE_SHIFT(line)))
+#define LINE_GET_COMPILER_PHASE(data, line)  (((data[line/CHAR_LINE] >> LINE_SHIFT(line)) & 7) - 1)
 
 /* Count of the hash collisions in the hash table */
 #ifdef GLOG_SHOW
 static unsigned int max_collisions = 0;
 #endif
 
+/* Helper macro to output the compiler phase */
+#define OUTPUT_COMPILER_PHASE(name, op, id, more) \
+    do { \
+        fprintf(fp, "\"%s\":[", name); \
+        lcount = 0; \
+        for (j = 1; j <= node->bmax; ++j) { \
+          if (LINE_IS_PRESENT_OR_COVERED(node->lines, j)) { \
+            if (LINE_GET_COMPILER_PHASE(node->lines, j) op id ) { \
+              if (lcount++) { \
+                fprintf(fp, ","); \
+              } \
+              fprintf(fp, "%d", j); \
+            } \
+          } \
+        } \
+        fprintf(fp, "]%s", more ? "," : ""); \
+    } while (0)
+
 /* Grow CoverNode bit set if necessary. */
 static void cover_node_ensure(CoverNode* node, int line);
 
 /* Add a node to the list of files */
-static CoverNode* add_get_node(CoverList *cover, const char *file);
+static CoverNode* add_get_node(CoverList* cover, const char* file);
 
 CoverList* cover_create(void) {
   CoverList* cover;
@@ -38,7 +61,7 @@ CoverList* cover_create(void) {
 
   cover->used = 0;
   cover->size = COVER_LIST_INITIAL_SIZE;
-  GMEM_NEWARR(cover->list, CoverNode**, COVER_LIST_INITIAL_SIZE, sizeof(CoverNode *));
+  GMEM_NEWARR(cover->list, CoverNode**, COVER_LIST_INITIAL_SIZE, sizeof(CoverNode*));
 
   return cover;
 }
@@ -60,18 +83,18 @@ void cover_destroy(CoverList* cover) {
     /* GLOG(("Destroying string [%s]", tmp->file)); */
     GMEM_DELSTR(tmp->file, -1);
     /* GLOG(("Destroying array [%p] with %d elements", tmp->lines, tmp->alen)); */
-    GMEM_DELARR(tmp->lines, unsigned char*, tmp->alen, sizeof(unsigned char*));
+    GMEM_DELARR(tmp->lines , unsigned char*,   tmp->alen, sizeof(unsigned char));
     /* GLOG(("Destroying node [%p]", tmp)); */
     GMEM_DEL(tmp, CoverNode*, sizeof(CoverNode));
     cover->list[i] = 0;
   }
 
   GLOG(("Destroying cover [%p]. Max run %d. Used: %d", cover, max_collisions, cover->used));
-  GMEM_DELARR(cover->list, CoverNode**, cover->size, sizeof(CoverNode *));
+  GMEM_DELARR(cover->list, CoverNode**, cover->size, sizeof(CoverNode*));
   GMEM_DEL(cover, CoverList*, sizeof(CoverList));
 }
 
-void cover_add_covered(CoverList* cover, const char* file, int line) {
+void cover_add_covered(CoverList* cover, const char* file, int line, int phase) {
   CoverNode* node = 0;
 
   if (file[0] == '(')
@@ -88,7 +111,7 @@ void cover_add_covered(CoverList* cover, const char* file, int line) {
   if (! LINE_IS_COVERED(node->lines, line)) {
     /* GLOG(("Adding line %d for [%s]", line, node->file)); */
     ++node->bcnt;
-    LINE_SET_COVERED(node->lines, line);
+    LINE_SET_COVERED(node->lines, line, phase);
   }
 }
 
@@ -150,9 +173,20 @@ void cover_dump(CoverList* cover, FILE* fp) {
         fprintf(fp, "%d", j);
       }
     }
-    fprintf(fp, "]}");
+
+    fprintf(fp, "],\"phases\":{");
+
+    OUTPUT_COMPILER_PHASE("BEGIN"   , <=, PERL_PHASE_START   , 1);
+    OUTPUT_COMPILER_PHASE("CHECK"   , ==, PERL_PHASE_CHECK   , 1);
+    OUTPUT_COMPILER_PHASE("INIT"    , ==, PERL_PHASE_INIT    , 1);
+    OUTPUT_COMPILER_PHASE("RUN"     , ==, PERL_PHASE_RUN     , 1);
+    OUTPUT_COMPILER_PHASE("END"     , ==, PERL_PHASE_END     , 1);
+    OUTPUT_COMPILER_PHASE("DESTRUCT", ==, PERL_PHASE_DESTRUCT, 0);
+
+    fprintf(fp, "}"); /* close the `phases` object */
   }
-  fprintf(fp, "}");
+  fprintf(fp, "}"); /* close the file object */
+  fprintf(fp, "}"); /* close the main object */
 }
 
 static void cover_node_ensure(CoverNode* node, int line) {
@@ -174,7 +208,7 @@ static void cover_node_ensure(CoverNode* node, int line) {
     /* GLOG(("Growing map for [%s] from %d to %d - %p", node->file, node->alen, size, node->lines)); */
 
     /* realloc will grow the data and keep all current values... */
-    GMEM_REALLOC(node->lines, unsigned char*, node->alen * sizeof(unsigned char*), size * sizeof(unsigned char*));
+    GMEM_REALLOC(node->lines,  unsigned char*,   node->alen * sizeof(unsigned char),   size * sizeof(unsigned char));
 
     /* ... but it will not initialise the new space to 0. */
     memset(node->lines + node->alen, 0, size - node->alen);
@@ -184,7 +218,7 @@ static void cover_node_ensure(CoverNode* node, int line) {
   }
 }
 
-static U32 find_pos(CoverNode **where, U32 hash, const char *file, int size) {
+static U32 find_pos(CoverNode** where, U32 hash, const char* file, int size) {
   U32 pos = hash % size;
 
 #ifdef GLOG_SHOW
@@ -210,15 +244,16 @@ static U32 find_pos(CoverNode **where, U32 hash, const char *file, int size) {
   return pos;
 }
 
-static CoverNode* add_get_node(CoverList *cover, const char *file) {
+static CoverNode* add_get_node(CoverList* cover, const char* file) {
   U32 hash, pos, i;
-  CoverNode *node = NULL, **new_list = NULL;
+  CoverNode* node = NULL;
+  CoverNode** new_list = NULL;
   ssize_t len = strlen(file);
 
   /* TODO: comment these magic numbers */
   /* TODO: move this enlargement code to a separate function */
   if (3 * cover->used > 2 * cover->size) {
-    GMEM_NEWARR(new_list, CoverNode**, cover->size * 2, sizeof(CoverNode *));
+    GMEM_NEWARR(new_list, CoverNode**, cover->size * 2, sizeof(CoverNode*));
     for (i = 0; i < cover->size; i++) {
       if (!cover->list[i]) {
         continue;
@@ -227,7 +262,7 @@ static CoverNode* add_get_node(CoverList *cover, const char *file) {
       new_list[pos] = cover->list[i];
     }
 
-    GMEM_DELARR(cover->list, CoverNode**, cover->size, sizeof(CoverNode *));
+    GMEM_DELARR(cover->list, CoverNode**, cover->size, sizeof(CoverNode*));
     cover->list = new_list;
     cover->size *= 2;
   }
@@ -244,9 +279,9 @@ static CoverNode* add_get_node(CoverList *cover, const char *file) {
   /* TODO: normalise name first? ./foo.pl, foo.pl, ../bar/foo.pl, etc. */
   int l = 0;
   GMEM_NEWSTR(node->file, file, -1, l);
-  node->lines = NULL;
-  node->hash = hash;
-  node->alen = node->bcnt = node->bmax = 0;
+  node->lines  = NULL;
+  node->hash   = hash;
+  node->alen   = node->bcnt = node->bmax = 0;
 
   ++cover->used;
   cover->list[pos] = node;
